local Object = require("lib.classic")
local TileMap = Object:extend()

function TileMap:new(tileset, path)
    self.layersByName = {}
    self.layerNames = {}
    self.tileset = tileset
    self.spriteBatch = love.graphics.newSpriteBatch(self.tileset.texture, 1000, "stream")
    self.width = 0
    self.height = 0
    self.tilewidth = 0
    self.tileheight = 0
    self.objects = {}
    self.objectsById = {}
    
    local fn = love.filesystem.load(path)
    self:load(fn())
end

function TileMap:load(tiledMap)
    self.width = tiledMap.width 
    self.height = tiledMap.height
    self.tilewidth = tiledMap.tilewidth
    self.tileheight = tiledMap.tileheight

    for k,v in ipairs(tiledMap.layers)  do
        if (v.type == "objectgroup") then
            for _,obj in ipairs(v.objects) do
                table.insert(self.objects, obj)
                self.objectsById[obj.id] = obj
            end
        end

        if (v.type == "tilelayer") then
            local layer = {}
            layer.name = v.name
            
            layer.properties = v.properties
            layer.parallaxx = v.parallaxx or 1
            layer.parallaxy = v.parallaxy or 1
            layer.data = {}
            for i,t in ipairs(v.data) do    
                    local ti = i-1
                    local x = (ti % tiledMap.width) * tiledMap.tilewidth
                    local y = (math.floor(ti / tiledMap.width)) * tiledMap.tileheight
    
                    local tile = {}
                    tile.tile = true
                    tile.id = t
                    tile.x = x
                    tile.y = y
                    tile.position = {x = x, y = y}
                    tile.w = tiledMap.tilewidth
                    tile.h = tiledMap.tileheight
    
                    table.insert(layer.data,i, tile)
            end
            self.layersByName[layer.name] = layer
            table.insert(self.layerNames, layer.name)
        end
    end
end

function TileMap:fromWorld(x, y)
    return math.floor(x / self.tilewidth), math.floor(y /self.tileheight)
end

local nullTile = {
    id = -1
}
function TileMap:getTileFromWorld(layer, x, y)
    local tx, ty = self:fromWorld(x, y)
    if tx < 0 or tx > self.width - 1 or ty < 0 or ty > self.height - 1 then
        return nullTile
    end
    local i = (ty * self.width) + tx
    return self.layersByName[layer].data[i +1]
end

function TileMap:toWorld(x, y)
    return x * self.tilewidth, y * self.tileheight
end

function TileMap:drawLayer(layerName,x,y,w,h)
    if not x then x = 0 end
    if not y then y = 0 end
    if not w then w = self.width * self.tilewidth end
    if not h then h = self.height * self.tileheight end


    local startX, startY = self:fromWorld(x, y)
    local endX, endY = self:fromWorld(x + w, y + h)

    self.spriteBatch:clear()

    local layer = self.layersByName[layerName]
    
    for ty = startY-1,endY+1 do
        for tx = startX-1,endX+1 do
            if (tx >= 0 and tx < self.width and ty >= 0 and ty < self.height) then 
                local index = (ty * self.width) + tx
            
                local t = layer.data[index+1]
                local q = self.tileset:get(t.id)
                if (q ~= nil) then
                    self.spriteBatch:add(q, math.floor(t.x) * DRAW_SCALE, math.floor(t.y) * DRAW_SCALE, 0, DRAW_SCALE, DRAW_SCALE)
                end
            end
            
        end
    end
    love.graphics.draw(self.spriteBatch)
end

function TileMap:draw(x,y,w,h,layers)
    if not layers then layers = self.layerNames end

    for _,layer in ipairs(layers) do
        self:drawLayer(layer,x,y,w,h)
    end

end

return TileMap