local Object = require("lib.classic")
local Sensors = require("framework.sensors")
local Entity = require("framework.entity")
local StateMachine = require("framework.state_machine")
local Physics = require("game.physics")

local function getHorizontalInput()
    return Input:actionStrength("right") - Input:actionStrength("left")
end

local function isJumpRequested(player)
    return player.jumpBuffer > 0 
end

local function isAttackRequested(player)
    return Input:isJustPressed("attack")
end


local function canWallGrab(player, dir, tilemap, toCheck)
    local sensors = player.sensors

    local pos = player.position

    sensors:update(pos.x+dir, pos.y, tilemap, toCheck) 
    local total = 0
    local totalColliding = 0
    local allColliding = true
    for _,s in  ipairs(toCheck) do
        if s.colliding then
            totalColliding = totalColliding + 1
        end
        total = total + 1
    end

    return totalColliding >= 2 
end


local PlayerAnimations = {
    idle = {
        {id=0, ox=8, oy=9, t=30},
    },
    running = {
        {id=1, ox=8, oy=9, t=4},
        {id=2, ox=8, oy=9, t=4},
        {id=3, ox=8, oy=9, t=4},
        {id=4, ox=8, oy=9, t=4},
        {id=3, ox=8, oy=9, t=4},
        {id=2, ox=8, oy=9, t=4},
        
    },
    jumping = {
        {id=5, ox=8, oy=9, t=9},
        {id=6, ox=8, oy=9, t=3000},
    }, 
    falling = {
        {id=7, ox=8, oy=9, t=3000},
    }, 
    attacking_stood = {
        {id=8, ox=8, oy=9, t=3},
        {id=9, ox=8, oy=9, t=3000},
    },
    attacking_airborne = {
        {id=10, ox=8, oy=9, t=3},
        {id=11, ox=8, oy=9, t=3000},
    },
    climbing = {
        {id=12, ox=8, oy=9, t=3000},
    },
    damaged = {
        {id=1, ox=8, oy=9, t=5},
    },
    dead = {
        {id=13, ox=8, oy=9, t=3},
        {id=14, ox=8, oy=9, t=3},
        {id=15, ox=8, oy=9, t=3},
        {id=16, ox=8, oy=9, t=3},
        {id=17, ox=8, oy=9, t=3000},
    }
}

local PlayerStates = {
    idle = {
        enter = function(sm, player)
            player.velocity.x = 0
            player.velocity.y = 0
            player.inAir = false
            player:playAnimation(PlayerAnimations.idle)

            -- immediately switch state if currently running
            if getHorizontalInput() ~= 0 then
                sm:switch("running", player)
            end
        end,
        
        update = function(sm, player, gameLogic, dt)
            Physics.applyGravity(player)
            local result = Physics.applyMovement(player, gameLogic)
            if bit.band(result, COLLIDE_DOWN) == 0 then
                sm:switch("falling", player)
            elseif getHorizontalInput() ~= 0 then
                sm:switch("running", player)
            elseif isJumpRequested(player) then
                sm:switch("jumping", player, player.stats.jumpVelocity)
            elseif isAttackRequested(player) then
                sm:switch("attacking_stood", player, gameLogic)
            end
        end
    },

    falling = {
        enter = function(sm, player, jumpGrace)
            player.inAir = true
            player.jumpGrace = jumpGrace or 0
            player:playAnimation(PlayerAnimations.falling)
        end,

        exit = function()
            
        end,
        
        update = function(sm, player, gameLogic, dt)
            if player.jumpGrace > 0 and isJumpRequested(player) then
                sm:switch("jumping", player, player.stats.jumpVelocity)
            else
                player.jumpGrace = math.max(0, player.jumpGrace - 1)
                Physics.applyGravity(player)
                local hInput = getHorizontalInput()
                
                local speed = hInput + player.dir == 0  and player.stats.groundSpeed * 0.4 or player.stats.groundSpeed
                player.velocity.x = hInput * speed
                

                local y = player.velocity.y
                local result = Physics.applyMovement(player, gameLogic)

                
                if bit.band(result, COLLIDE_RIGHT+COLLIDE_LEFT) ~= 0 then
                   
                    if canWallGrab(player, -1, gameLogic.tilemap, player.sensors.left) then
                        player.dir = -1
                        sm:switch("climbing", player)
                    elseif canWallGrab(player, 1, gameLogic.tilemap, player.sensors.right)  then
                        player.dir = 1
                        sm:switch("climbing", player)
                    elseif player.velocity.y == 0 then
                        sm:switch("idle", player)
                    end
                elseif isAttackRequested(player) then
                    sm:switch("attacking_airborne", player, gameLogic, hInput ~= 0 and hInput or player.dir)
                elseif player.velocity.y == 0 then
                    sm:switch("idle", player)
                    Assets.sfx.land:play()
                end
            end
        end
    },

    jumping = {
        enter = function(sm, player, jumpVelocity, ignoreInputT)
            player.t = 0
            player.inAir = true
            player.velocity.y = jumpVelocity 
            player.ignoreInputT = ignoreInputT or 0
            player:playAnimation(PlayerAnimations.jumping)
            Assets.sfx.jump:play()
        end,
        
        update = function(sm, player, gameLogic, dt)
            local hInput = getHorizontalInput()
            if player.t < player.ignoreInputT then
                hInput = player.dir
            end
            local speed = hInput + player.dir == 0  and player.stats.groundSpeed * 0.4 or player.stats.groundSpeed
            player.velocity.x = hInput * speed
            

            Physics.applyGravity(player)
            local result = Physics.applyMovement(player, gameLogic)
            
            if bit.band(result, COLLIDE_RIGHT+COLLIDE_LEFT) ~= 0 then
                if canWallGrab(player, -1, gameLogic.tilemap, player.sensors.left) then
                    player.dir = -1
                    sm:switch("climbing", player)
                elseif canWallGrab(player, 1, gameLogic.tilemap, player.sensors.right)  then
                    player.dir = 1
                    sm:switch("climbing", player)
                end
            elseif isAttackRequested(player) then
                sm:switch("attacking_airborne", player, gameLogic, hInput ~= 0 and hInput or player.dir)
            elseif player.velocity.y >= 0 then
                sm:switch("falling", player)
            end
        end
    },

    running = {
        enter = function(sm, player)
            player.inAir = false
            player:playAnimation(PlayerAnimations.running)
        end,

        update = function(sm, player, gameLogic, dt)
            Physics.applyGravity(player)
            local hInput = getHorizontalInput()
            
            player.velocity.x = hInput * player.stats.groundSpeed
            if hInput ~= 0 then
                player.dir = hInput
            end

            local result = Physics.applyMovement(player, gameLogic)
            if isJumpRequested(player) then
                sm:switch("jumping", player, player.stats.jumpVelocity)
            elseif isAttackRequested(player) then
                sm:switch("attacking_stood", player, gameLogic)
            elseif player.velocity.y > 0 then
                sm:switch("falling", player, player.stats.jumpGrace)
            elseif hInput == 0 then
                sm:switch("idle", player)
            end
        end
    },

    climbing = {
        enter = function(sm, player)
            player.inAir = false
            player.velocity.x = 0
            player.velocity.y = 0
            player:playAnimation(PlayerAnimations.climbing)
            Assets.sfx.land:play()
        end,

        update = function(sm, player, gameLogic)
            local hInput = getHorizontalInput()

            if Input:isJustPressed("jump") then
                player.dir = player.dir * -1
                sm:switch("jumping", player, player.stats.jumpVelocity * 0.9, 6 )
            end
        end
    },

    attacking_stood = {
        enter = function(sm, player, gameLogic)
            player.t = 0 
            player.velocity.x = 0
            local x, y = player.position.x, player.position.y
            local duration = player.stats.attackDuration
            gameLogic.swordAttack:attack(x, y, duration, player.dir)
            player:playAnimation(PlayerAnimations.attacking_stood)
            Assets.sfx.slash:play()
        end,

        update = function(sm, player, gameLogic)
            -- might not be necessary
            Physics.applyGravity(player)
            Physics.applyMovement(player, gameLogic)
            if player.t == player.stats.attackDuration then
                sm:switch("idle", player, gameLogic)
            elseif isJumpRequested(player) and player.t >= player.stats.attackCancelTime then
                gameLogic.swordAttack:cancel()
                sm:switch("jumping", player, player.stats.jumpVelocity)
            end
        end
    },

    attacking_airborne = {
        enter = function(sm, player, gameLogic, dir)
            player.t = 0 
            player.dir = dir or player.dir
            local x, y = player.position.x, player.position.y
            local duration = player.stats.attackDuration
            gameLogic.swordAttack:attack(x, y, duration, player.dir)
            player:playAnimation(PlayerAnimations.attacking_airborne)
            Assets.sfx.slash:play()
        end,

        update = function(sm, player, gameLogic)
            Physics.applyGravity(player)
            local hInput = getHorizontalInput()
            local speed = hInput + player.dir == 0  and player.stats.groundSpeed * 0.6 or player.stats.groundSpeed
            player.velocity.x = hInput * speed
            
            local result = Physics.applyMovement(player, gameLogic)
            if player.t == player.stats.attackDuration then
                sm:switch("falling", player)
            elseif bit.band(result, COLLIDE_RIGHT+COLLIDE_LEFT) ~= 0 then
                if canWallGrab(player, -1, gameLogic.tilemap, player.sensors.left) then
                    gameLogic.swordAttack:cancel()
                    player.dir = -1
                    sm:switch("climbing", player)
                elseif canWallGrab(player, 1, gameLogic.tilemap, player.sensors.right)  then
                    gameLogic.swordAttack:cancel()
                    player.dir = 1
                    sm:switch("climbing", player)
                end
            -- elseif isJumpRequested(player) and player.t <= 4 then 
            --     -- this was a bug but I thought it was funny so left it in and made it timed
            --     gameLogic.swordAttack:cancel()
            --     sm:switch("jumping", player, player.stats.jumpVelocity)
            elseif player.velocity.y == 0 then 
                gameLogic.swordAttack:cancel()
                sm:switch("idle", player)
            end
        end
    },

    damaged = {
        enter = function(sm, player, dir)
            player.t = 0
            player.velocity.y = player.stats.jumpVelocity * 0.6
            player.velocity.x = dir * player.stats.groundSpeed * 0.6
            player:playAnimation(PlayerAnimations.damaged)
            Assets.sfx.playerHit:play()
        end,

        update = function(sm, player, gameLogic)
            if player.t > 20 then
                if player.velocity.y ~= 0 then
                    sm:switch("falling", player)
                else
                    sm:switch("idle", player)
                end
                return
            end
            Physics.applyGravity(player)
            local result = Physics.applyMovement(player, gameLogic)

            if bit.band(result, COLLIDE_RIGHT+COLLIDE_LEFT) ~= 0 then
                if canWallGrab(player, -1, gameLogic.tilemap, player.sensors.left) then
                    gameLogic.swordAttack:cancel()
                    player.dir = -1
                    sm:switch("climbing", player)
                elseif canWallGrab(player, 1, gameLogic.tilemap, player.sensors.right)  then
                    gameLogic.swordAttack:cancel()
                    player.dir = 1
                    sm:switch("climbing", player)
                end
            elseif bit.band(result, COLLIDE_DOWN) ~= 0 then
                sm:switch("idle", player)
            end
 
        end
    },

    dead = {
        enter = function(sm, player)
            player:playAnimation(PlayerAnimations.dead)
            Assets.sfx.playerDie:play()
        end,

        update = function()

        end
    }
}



local function updatePlayer(player, gameLogic, dt)
    player.t = (player.t + 1) % 256 -- general purpose timer used for all sorts of hacky nonsense
    player.jumpBuffer = math.max(0, player.jumpBuffer-1)

    if Input:isJustPressed("jump") then
        player.jumpBuffer = player.stats.jumpBufferTime
    end

    player.stateMachine:update(player, gameLogic, dt)
    player.flipH = player.dir < 0

    local pos = player.position
    local levelWidth = gameLogic.tilemap.width * gameLogic.tilemap.tilewidth
    pos.x = math.min(levelWidth-player.extents.w, math.max(player.extents.w, pos.x))

    if pos.y < 0 then
        pos.y = 0
        if player.velocity.y < 0 then
            player.velocity.y = 0
        end
    end

    if player.stats.invincibility > 0 then
        player.visible = not player.visible
    else
        player.visible = true
    end
end

return function(x, y)
    local stats = {}
    stats.jumpVelocity, stats.gravity = kinematicJump(3.5*CELL_SIZE, 0.25)
    stats.groundSpeed = 0.6
    stats.maxFallSpeed = 2.5
    stats.maxHp = 4
    stats.hp = stats.maxHp
    stats.invincibility = 0
    stats.jumpBufferTime = 6
    stats.jumpGrace = 6
    stats.attackDuration = 14
    stats.attackCancelTime = 6
    

    return Entity(x, y, {
        dir = 1, 
        t = 0,
        stats = stats,
        stateMachine = StateMachine(PlayerStates, "idle"),
        extents = {w=2, h=3},
        atlas = Assets.atlases.player,

        sensors = Sensors({
            left = {
                {x=-2, y=-3},   
                {x=-2, y=-1},   
                {x=-2, y=4},   
            },
            right = {
                {x=-0, y=-3},   
                {x=-0, y=-1}, 
                {x=-0, y=4},   
            },
            up = {
                {x=-2, y=-3},   
                {x=-0, y=-3},   
            },
            down = {
                {x=-2, y=4},   
                {x=-0, y=4},   
            },
        }),
    
        jumpBuffer = 0,
        init = function(player, gameLogic)
            player.stateMachine:switch("idle", player)
        end,
    
        update = updatePlayer,
        onTakeDamage = function(player, obj)
            local dir = 0
            local dx = player.position.x - obj.position.x 
            if dx < 0 then
                dir = -1
            else
                dir = 1
            end
            player.stateMachine:switch("damaged", player, dir)
        end,

        onDeath = function(player, obj)
            player.stateMachine:switch("dead", player)
        end,

        isDead = function(player)
            return player.stats.hp <= 0
        end
    })
end