local Object = require("lib.classic")
local Camera = require("framework.camera")
local TileMap = require("framework.tilemap")
local Player = require("game.entities.player")
local PlayerSword = require("game.entities.player_sword")

local Entity = require("framework.entity")

local GameLogic = Object:extend()

local Animations = {
    flash = {
        {id=0, ox=4, oy=4, t=1},
        {id=1, ox=4, oy=4, t=1},
        {id=2, ox=4, oy=4, t=1},
        {id=3, ox=4, oy=4, t=1},
        {id=4, ox=4, oy=4, t=1},
        {id=5, ox=4, oy=4, t=1},
        {id=6, ox=4, oy=4, t=3000},
    }
}


local function tablerm(list, obj)
    for i=1,#list do
        if list[i] == obj then
            table.remove(list, i)
            return
        end
    end
end

local function tableins(list, value)
    local index = #list+1
    for i=1,#list do
        if list[i] == nil then
            index = i
        end
    end
    list[index] = value
end

local function smoothPosition(alpha, malpha, position)
    position.sx = (alpha * position.x) + (malpha * position.px)
    position.sy = (alpha * position.y) + (malpha * position.py)
end

local function resetSmoothPosition(p)
    p.sx = p.x
    p.px = p.x
    p.sy = p.y
    p.py = p.y
end

local function draw(obj)
    if obj.visible == false then return end

    local pos = obj.position
    local atlas = obj.atlas
    local frame = obj.animation[obj.frame]
    local flipH = obj.flipH and -1 or 1
    local flipV = obj.flipV and -1 or 1

    if pos and atlas and frame then
        atlas:draw(frame.id, pos.sx or pos.x, pos.sy or pos.y, frame.ox, frame.oy, 0, flipH, flipV)
    end
end


local function update(obj, gameLogic, dt)
    if obj == nil or not obj.active then return end
    local fn = obj.update
    if fn then
        fn(obj, gameLogic, dt)
    end

    -- update current animation
    obj.frameTimer = obj.frameTimer + 1
    local currentFrame = obj.animation[obj.frame]
    if currentFrame and obj.frameTimer > currentFrame.t then
        local newFrame = obj.frame + 1
        if newFrame > #obj.animation then
            newFrame = 1
        end
        obj.frame = newFrame
        obj.frameTimer = 0
    end
end

local function updateEffect(obj)
    obj.frameTimer = obj.frameTimer + 1
    local currentFrame = obj.animation[obj.frame]
    if currentFrame and obj.frameTimer > currentFrame.t then
        local newFrame = obj.frame + 1
        if newFrame > #obj.animation then
            return true
        end
        obj.frame = newFrame
        obj.frameTimer = 0
    end
    return false
end

local function updateEntities(entities, ...)
    local toRemove = {}
    for i, obj in ipairs(entities) do
        update(obj, ...)
        if obj.destroyed then
            table.insert(toRemove, obj)
        end
    end

    for _,i in ipairs(toRemove) do
        tablerm(entities, i)
    end
end

function processList(l, fn, ...)
    local toRemove = {}
    for i=1,#l do
        
            local remove = fn(l[i], ...)
            if remove == true then 
                table.insert(toRemove, l[i])
            end
        
    end

    if #toRemove > 0 then
        
        for i=1,#toRemove do
            tablerm(l,toRemove[i])
        end
        
    end
end

function effect(x, y, prot)
    return {
        atlas = prot.atlas,
        animation = prot.animation,
        frame = 1,
        frameTimer = 0,
        position = {x = x, y = y}
    }
end

flashEffect = {
    atlas = Assets.atlases.flash,
    animation = {
        {id=0, ox=4, oy=4, t=1},
        {id=1, ox=4, oy=4, t=1},
        {id=2, ox=4, oy=4, t=1},
        {id=3, ox=4, oy=4, t=1},
        {id=4, ox=4, oy=4, t=1},
        {id=5, ox=4, oy=4, t=1},
    }
}

function GameLogic:new(levelNumber, session)
    self.session = session
    self.levelNumber = levelNumber
    local level = Levels[levelNumber]
    level:load()
    self.level = level
    self.tilemap = level.tilemap
    self.camera = Camera(DRAW_WIDTH, DRAW_HEIGHT)
    local cam = self.camera
    cam.min_x = 0
    cam.min_y = 0
    cam.max_x = self.tilemap.width * self.tilemap.tilewidth
    cam.max_y = cam.height

    -- anything that can damage enemies
    self.playerAttacks = {}
    self.player = Player(unpack(self.level.playerSpawn))
    self.swordAttack = PlayerSword(self.player)

    -- sword attack is never removed
    table.insert(self.playerAttacks, self.swordAttack)

    self.spawners = {}
    -- anything that can damage the player
    self.enemies = {}
    self.enemiesToRemove = {}
    self.effects = {}
    SoundManager:playMusic(level.music)
    
    self.t = 0
    self.gt = 0

    self.level:start(self)
end

function GameLogic:addEnemy(enemy)
    for i=1,#self.enemies do
        if enemy.sid ~= null and self.enemies[i].sid == enemy.sid then
            return
        end
    end
    table.insert(self.enemies, enemy)
end

function GameLogic:removeEnemy(enemy)
    tablerm(self.enemies, enemy)
end

function GameLogic:update(dt)
    self.gt = self.gt + 1
    self.t = (self.t + 1) % 256

    if self.player:isDead() and self.t > 120 then
        
        if self.session.lives > 0 then
            self.session.lives = self.session.lives - 1
            Events.switchState(States.Game, self.levelNumber, self.session)
        else
            Events.switchState(States.GameOver)
        end
    end

    self.level:update(self, dt)
    update(self.player, self,dt)
    self.camera:setPosition(self.player.position.x + 4, self.player.position.y)

    updateEntities(self.playerAttacks, self, dt)

    updateEntities(self.enemies, self, dt)
    processList(self.enemies, function(obj)
        if obj.invincibility then
            obj.invincibility = math.max(0, obj.invincibility - 1)
        end
        local cam = self.camera

        local x,y,w,h = cam.pos.x-cam.half_width, cam.pos.y-cam.half_height, cam.width, cam.height

        if obj.position.x < x-16 or obj.position.x > x + w + 16 or obj.position.y > 100 then
            return true
        end
        return false
    end)


    -- damage enemies
    local attacksToRemove = {}
    local toRemove = {}
    for i, obj in ipairs(self.playerAttacks) do
        if obj.active  then
            for ei, enemy in ipairs(self.enemies) do
                if enemy.static ~= true and enemy.active and enemy.stats.hp > 0 and obj:overlapsWith(enemy) and (enemy.invincibility == 0 or enemy.invincibility == nil) then
                    enemy.stats.hp = enemy.stats.hp - 1
                    enemy.invincibility = 120
                    if enemy.onTakeDamage then
                        enemy:onTakeDamage(obj)     
                    end
                    if obj ~= self.swordAttack then
                        table.insert(attacksToRemove, obj)
                    end
                    if enemy.stats.hp <= 0 then
                        local killEnemy = true
                        if enemy.onDeath then 
                            if enemy:onDeath(self) == true then
                                killEnemy = false
                            end

                        end
                        if killEnemy then
                            self:kill(enemy)

                        end
                        --self:kill(enemy)
                    end
                end
            end
        end
    end

    for _,i in pairs(attacksToRemove) do
        tablerm(self.playerAttacks, i)
    end

    for _,enemy in ipairs(self.enemiesToRemove) do
        -- table.insert(toRemove, enemy)
        table.insert(self.effects, effect(enemy.position.x, enemy.position.y, flashEffect))
        Assets.sfx.enemyDie:play()
        tablerm(self.enemies, enemy)
    end
    self.enemiesToRemove = {}

    for _,i in ipairs(toRemove) do
        tablerm(self.enemies, i)
    end


    -- damage the player
    local player = self.player
    player.stats.invincibility = math.max(0, player.stats.invincibility - 1)
    local pd = player:isDead()
    if player.stats.hp > 0 and  player.stats.invincibility == 0 then
        for i, obj in ipairs(self.enemies) do
            local w = player.extents.w + obj.extents.w
            local h = player.extents.h + obj.extents.h

            if pd == false and player.stats.invincibility == 0 and player:overlapsWith(obj) then
                local damage = obj.damage or 1
                player.stats.hp = player.stats.hp - damage
                if player.stats.hp <= 0 then
                    player:onDeath(obj)
                    self.t = 0
                    SoundManager:stopMusic()
                    
                else
                    player:onTakeDamage(obj)
                    player.stats.invincibility = 90
                end
            end
        end
    end
    
    if self.player:isDead() == false and self.player.position.y > 80 then
        player.stats.hp = 0
        player:onDeath()
        SoundManager:stopMusic()
        self.t = 0
    end

    local effectsToRemove = {}
    for _, effect in ipairs(self.effects) do
        if updateEffect(effect) then
            table.insert(effectsToRemove, effect)
        end
    end
    for _, e in ipairs(effectsToRemove) do
        tablerm(self.effects, e)
    end
    
    if self.level:checkEndCondition(self)  then
        local nextLevel = self.levelNumber + 1
        SoundManager:stopMusic()
        if nextLevel > #Levels then
            Events.switchState(States.GameBeat)
        else
            Events.switchState(States.LevelIntro, self.levelNumber + 1, self.session)
        end
    end
end

function GameLogic:kill(enemy)
    table.insert(self.enemiesToRemove, enemy)
end

function GameLogic:draw()
    love.graphics.clear(158/255,183/255,149/255,1)

    self.level:draw(self)
    
    self.camera:attach()
    for i, obj in ipairs(self.enemies) do
        draw(obj)
    end
    draw(self.player)
    
    for _,obj in ipairs(self.effects) do
        draw(obj)
    end

    self.level:drawFront(self)     
    
    self.camera:detach()
    
    love.graphics.setColor(25/255,25/255,27/255,1)
    love.graphics.rectangle("fill",0,0,DRAW_WIDTH*DRAW_SCALE,6*DRAW_SCALE)
    love.graphics.setColor(1,1,1,1)
    Assets.hud.heart:draw(0, 0,0)
    for i = 0,self.player.stats.maxHp-1 do
        local x = i * 2
        Assets.hud.energy:draw((i+1)<=self.player.stats.hp and 0 or 1, 6 + x, 0)
    end

    Assets.hud.life:draw(0,DRAW_WIDTH-6,0)
    local xs = DRAW_WIDTH - 8
    for i = 0, 2 do
        if i+1<=self.session.lives then
            local x = xs - (i * 2)
            Assets.hud.energy:draw(0, x, 0)
        end
    end

end

function GameLogic:resetSmoothStates()
    self.camera.smoothX = self.camera.pos.x
    self.camera.smoothY = self.camera.pos.y
    self.camera.prevX = self.camera.pos.x
    self.camera.prevY = self.camera.pos.y

    resetSmoothPosition(self.player.position)
    for i,obj in ipairs(self.enemies) do
        resetSmoothPosition(obj.position)
    end
end

function GameLogic:smoothStates(alpha)
    -- if alpha < 0.01 then
    --     alpha = 0
    -- end
    local malpha = 1.0 - alpha

    self.camera.smoothX = ((alpha * self.camera.pos.x) + (malpha * self.camera.prevX))
    self.camera.smoothY = ((alpha * self.camera.pos.y) + (malpha * self.camera.prevY))

    smoothPosition(alpha, malpha, self.player.position)
    for i,obj in ipairs(self.enemies) do
        smoothPosition(alpha, malpha, obj.position)
    end
end



return GameLogic